#ifndef SHA1_T_H
#define SHA1_T_H

#include <assert.h>
#include <stdint.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

/* A SHA1 class.
 * This class both represents the SHA1 and can calculate it from a supplied buffer.
 *
 * This instrumented sha1 keeps track of how many times it has been called and 
 * has a bypass for calculating SHA1 of all NULLs.
 * 
 * Uses the OpenSSL EVP class
 */

#ifdef HAVE_OPENSSL_HMAC_H
#include <openssl/hmac.h>
#else
#error Requires OpenSSL hmac.h
#endif

class hashbase_t {
 public:
    /* Static function to determine if something is zero */
    static bool iszero(const u_char *buf,size_t bufsize){
	for(u_int i=0;i<bufsize;i++){
	    if(buf[i]!=0) return false;
	}
	return true;
    }
 hashbase_t(const EVP_MD *md_):md(md_),initialized(false),finalized(false){
	strcpy(lovely,"hashbase_t");
    }
    ~hashbase_t(){
	clear();			/* be sure to clear allocated memory */
    }

    char lovely[64];
    mutable EVP_MD_CTX mdctx;			// the context for computing the value
    const EVP_MD *md;
    bool initialized;				/* has the context been initialized? */
    mutable bool finalized;
    mutable u_char final_hash[EVP_MAX_MD_SIZE]; /* where we put it */
    size_t size(){				/* return the size of the hash */
	return EVP_MD_size(md);
    }
    void init(){
	if(initialized==false){
	    EVP_MD_CTX_init(&mdctx);
	    EVP_DigestInit_ex(&mdctx, md, NULL);
	    initialized = true;
	    finalized = false;
	}
    }

    void update(const u_char *buf,size_t bufsize){
	if(!initialized) init();
	if(finalized){
	    fprintf(stderr,"sha1_t::update called after finalized\n");
	    exit(1);
	}
	EVP_DigestUpdate(&mdctx,buf,bufsize);
    }
    /** Return to original state. After this you can update but that's it */
    void clear(){
	if(initialized){
	    EVP_MD_CTX_cleanup(&mdctx);
	    initialized = false;
	}
    }
    /** Return the final value of the hash (finalize if necessary)
     * It's not really const, but we claim that we are so that we work with STL maps.
     */
    const u_char *final() const {
	if(!finalized){
	    if(!initialized){
		fprintf(stderr,"sha1_t::final called before init!n");
		exit(1);
	    }
	    unsigned int len = sizeof(final_hash);
	    EVP_DigestFinal(&mdctx,final_hash,&len);
	    finalized = true;
	}
	return final_hash;
    }

    /* hexbuf returns the hex representation of the hash */
    const char *hex(char *hexbuf,size_t bufsize) {
	const char *hexbuf_start = hexbuf;
	final();			/* be sure we have a finalized version */
	for(int i=0;i<EVP_MD_size(md) && bufsize>3;i++){
	    snprintf(hexbuf,bufsize,"%02x",final_hash[i]);
	    hexbuf  += 2;
	    bufsize -= 2;
	    
	}
	return hexbuf_start;
    }

    /* this implementation doesn't change the hash but requires that it be finalized*/
    const char *hex(char *hexbuf,size_t bufsize) const {
	assert(finalized);
	const char *hexbuf_start = hexbuf;
	for(int i=0;i<EVP_MD_size(md) && bufsize>3;i++){
	    snprintf(hexbuf,bufsize,"%02x",final_hash[i]);
	    hexbuf  += 2;
	    bufsize -= 2;
	    
	}
	return hexbuf_start;
    }

    /* set the hash from a hex string; returns the number of bits that were used.
     */
    int sethex(const char *hex) {
	clear();			
	finalized = true;
	return nsrl_hex2bin(final_hash,sizeof(final_hash),hex);
    }
};

static uint64_t sha1_ctr=0;			// number of SHA1 operations performed
static u_char sha1_512nulls[20];		/* hash of a null block, as an optimization */
class sha1_t:public hashbase_t {
 public:
    /* Create a hasher that can only accumulate */
 sha1_t():hashbase_t(EVP_sha1()){
    }
    /* Create a hash from a binary value. It can only be used for comparision */
 sha1_t(const u_char provided_hash[20]):hashbase_t(EVP_sha1()){
	initialized = false;
	finalized = true;		/* because we just have the value */
	memcpy(final_hash,provided_hash,sizeof(provided_hash));
    }

    /** Compute a sha1 from a buffer */
    sha1_t(const u_char *buf,size_t bufsize):hashbase_t(EVP_sha1()){
	/* First time through find the SHA1 of 512 NULLs */
	if(sha1_ctr==0){
	    u_char b2[512];
	    EVP_MD_CTX ctx2;
	    unsigned int len = sizeof(sha1_512nulls);
	    memset(b2,0,sizeof(b2));
	    EVP_MD_CTX_init(&ctx2);
	    EVP_DigestInit_ex(&ctx2, md, NULL);
	    EVP_DigestInit(&ctx2,md);
	    EVP_DigestUpdate(&ctx2,b2,sizeof(b2));
	    EVP_DigestFinal(&ctx2,sha1_512nulls,&len);
	    sha1_ctr++;
	}

	/* If the input is 512 bytes long and all NULLs, use our SHA1 of 512 NULLs */
	if(bufsize==512 && iszero(buf,bufsize)){
	    memcpy(final_hash,sha1_512nulls,sizeof(sha1_512nulls));
	    initialized = false;	/* never called init */
	    finalized = true;		/* but it is finalized! */
	    return;
	}
	if(initialized==0) init();
	update(buf,bufsize);
	final();
    }
    const u_char *final() const{			/* this version counts how many hashes have been finalized */
	sha1_ctr++;
	return hashbase_t::final();		/* super->final() */
    }
	
    /** Static method allocator */
    static sha1_t *new_from_hex(const char *hex){
	sha1_t *obj = new sha1_t();
	obj->sethex(hex);
	return obj;
    }

    /** Static method allocateor */
    static sha1_t *new_from_file(const char *fname){
	sha1_t *obj = new sha1_t();
	int fd = open(fname,O_RDONLY);
	if(fd<0) return 0;
	struct stat st;
	if(fstat(fd,&st)<0){
	    close(fd);
	    return 0;
	}
	const u_char *buf = (const u_char *)mmap(0,st.st_size,PROT_READ,MAP_FILE|MAP_SHARED,fd,0);
	if(buf==0){
	    close(fd);
	    return 0;
	}
	obj->update(buf,st.st_size);
	obj->final();
	munmap((void *)buf,st.st_size);
	close(fd);
	return obj;
    }

    /** Operator makes this class work with STL maps. Unfortunately, that means that they need to take
     * const, which means that they lie about the mutability change.
     */
    bool operator<(const sha1_t &s2) const 
    {
	return memcmp(this->final(),s2.final(), EVP_MD_size(md)) < 0;
    }
    bool operator==(const sha1_t &s2) const
    {
	return memcmp(this->final(),s2.final(), EVP_MD_size(md)) == 0;
    }
};


#endif
